from unittest import TestCase
from datetime import datetime

from src.aero_svo_api import utils


class TestUtils(TestCase):

    def test_format_date(self):
        date = datetime.fromisoformat('2020-10-19T12:11:10.343138+03:00')
        formatted_date = utils.format_date(date=date)
        self.assertEqual(formatted_date, '2020-10-19T12:11:10%2B03:00')





timetable = 'https://www.svo.aero/bitrix/timetable/'
terminals = 'https://www.svo.aero/bitrix/map/terminals/'
season_airports = 'https://www.svo.aero/bitrix/season/?locale=en'
season = 'https://www.svo.aero/bitrix/season/list_v2/?city=432&terminal=all&direction=departure&locale=en'
airlines = 'https://www.svo.aero/bitrix/airlines/?perPage=100&page=1&locale=ru'



from src.aero_svo_api.api import AsyncSvoApi, svo_logger
from src.aero_svo_api import urls
from src.aero_svo_api import models
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, Mock
from datetime import datetime, timedelta

from . import payload


class TestAsyncSvoApi(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.api = AsyncSvoApi()
        self.mock_response = AsyncMock(raise_for_status=Mock)
        self.api.session.get = AsyncMock(return_value=self.mock_response)

    async def asyncTearDown(self) -> None:
        await self.api.session.close()

    async def test_request(self):
        response = await self.api._request(urls.timetable, proxy='https://example.com', params={'param': 'val'})
        self.api.session.get.assert_called_with(urls.timetable, proxy='https://example.com', params={'param': 'val'})

    async def test_request_inject_session(self):
        json_mock = AsyncMock(return_value=payload.FlightPayload().model_dump())
        mock_session = AsyncMock(get=AsyncMock(return_value=Mock(json=json_mock)))

        response = await self.api.get_flight(flight_id=12345, _session=mock_session)
        mock_session.get.assert_called()

    async def test_session_injection_does_not_create_session(self):
        api = AsyncSvoApi()
        mock_session = AsyncMock(get=AsyncMock(return_value=AsyncMock(raise_for_status=Mock)))

        response = await api._request(urls.timetable, params={}, _session=mock_session)

        self.assertIsNone(api._session)

    async def test_context_manager(self):
        async with AsyncSvoApi() as api:
            self.assertFalse(api.session.closed)
        self.assertTrue(api.session.closed)

    async def test_get_schedule(self):
        data = {'items': [payload.FlightPayload().model_dump() for _ in range(10)]}
        self.mock_response.json.return_value = data

        schedule = await self.api.get_schedule(
            direction='arrival',
            date_start=datetime.now() - timedelta(hours=4),
            date_end=datetime.now(),
        )

        self.assertEqual(len(schedule.flights), 10)

    async def test_get_schedule_logging(self):
        self.mock_response.json.return_value = models.Schedule.model_construct(
            items=[
                payload.FlightPayload(i_id='string').model_dump(),      # invalid
                payload.FlightPayload().model_dump(),                   # valid
            ]
        ).model_dump(by_alias=True)

        with self.assertLogs(svo_logger, level='WARNING') as log:
            schedule = await self.api.get_schedule(
                direction='arrival',
                date_start=datetime.now() - timedelta(hours=4),
                date_end=datetime.now()
            )
            self.assertIn('validation error', log.output[0])
            self.assertEqual(len(schedule.flights), 1)

    async def test_get_flight(self):
        self.mock_response.json.return_value = payload.FlightPayload(i_id='12345').model_dump()
        flight = await self.api.get_flight(flight_id=12345)

        self.assertEqual(flight.id, 12345)




from datetime import datetime
from decimal import Decimal
from typing import Annotated, Literal
from pydantic import BaseModel, Field, model_validator, field_validator, AnyHttpUrl


class Base(BaseModel):
    @model_validator(mode='before')
    @classmethod
    def empty_str_to_none(cls, data: dict) -> dict:
        return {k: v if v != '' else None for k, v in data.items()}


class Aircraft(Base):
    id: int = Field(alias='aircraft_type_id')
    name: str = Field(alias='aircraft_type_name')


class Country(Base):
    name: str = Field(alias='country')
    region: str


class City(Base):
    name_en: str = Field(alias='city_eng')
    name_ru: str = Field(alias='city')
    timezone: str
    country: Country


class Airport(Base):
    id: int
    iata: str = Field(pattern=r'^[A-Z]{3}$')
    icao: str = Field(pattern=r'^[A-Z]{4}$')
    code_ru: str | None = Field(None, pattern=r'[A-ZА-Я]{3}', alias='rus')
    name: str = Field(alias='airport')
    name_ru: str = Field(alias='airport_rus')
    lat: Annotated[Decimal, Field(max_digits=9, decimal_places=6)] | None = None
    long: Annotated[Decimal, Field(max_digits=9, decimal_places=6)] | None = None
    city: City

    @model_validator(mode='before')
    @classmethod
    def nest_city(cls, data: dict) -> dict:
        if not isinstance(data.get('country'), dict):
            data['city'] = dict(
                city_eng=data.pop('city_eng'),
                city=data.pop('city'),
                timezone=data.pop('timezone'),
                country=dict(
                    country=data.pop('country'),
                    region=data.pop('region'),
                ),
            )

        return data


class Company(Base):
    code: str = Field(pattern='^[A-Z0-9]{2}$')
    name: str
    url_buy: AnyHttpUrl | None = Field(None, alias='onlineBuy')
    url_register: AnyHttpUrl | None = Field(None, alias='onlineRegister')


class Flight(Base):
    id: int = Field(alias='i_id')
    direction: Literal['arrival', 'departure'] = Field(alias='ad')
    company: Company = Field(alias='co')
    number: str = Field(alias='flt')
    date: datetime = Field(alias='dat')
    mar1: Airport
    mar2: Airport
    mar3: Airport | None = None
    mar4: Airport | None = None
    mar5: Airport | None = None
    aircraft: Aircraft
    main_id: int | None = Field(None, alias='main_flight')
    way_time: int
    # check-in
    chin_start: datetime | None = Field(None, alias='t_chin_start')
    chin_end: datetime | None = Field(None, alias='t_chin_finish')
    chin_start_et: datetime | None = Field(None, alias='estimated_chin_start')
    chin_end_et: datetime | None = Field(None, alias='estimated_chin_finish')
    # boarding
    boarding_start: datetime | None = Field(None, alias='t_boarding_start')
    boarding_end: datetime | None = Field(None, alias='t_boarding_finish')
    gate_id: str | None = None
    gate_id_prev: str | None = Field(None, alias='old_gate_id')
    # terminal
    term_local: str | None = Field(None, alias='term')
    term_local_prev: str | None = Field(None, alias='old_term')
    # bag belt
    bbel_id: str | None = None
    bbel_id_prev: str | None = Field(None, alias='old_bbel_id')
    bbel_start: datetime | None = None
    bbel_start_et: datetime | None = Field(None, alias='estimated_bag_start')
    bbel_end: datetime | None = Field(None, alias='bbel_finish')
    # schedule
    sked_local: datetime | None= Field(None, alias='t_st')
    sked_other: datetime | None = Field(None, alias='t_st_mar')
    # landing / takeoff
    at_local: datetime | None = Field(None, alias='t_at')
    at_local_et: datetime | None = Field(None, alias='t_et')
    at_other: datetime | None = Field(None, alias='t_at_mar')
    at_other_et: datetime | None = Field(None, alias='marArrivalEt')
    takeoff_et: datetime | None = Field(None, alias='fplTime')
    # departure / arrival to pk
    otpr: datetime | None = Field(None, alias='t_otpr')
    prb: datetime | None = Field(None, alias='t_prb')
    # status
    status_id: int | None = None
    status_code: int | None = None

    @field_validator('direction', mode='before')
    @classmethod
    def convert_direction(cls, value):
        return {'A': 'arrival', 'D': 'departure'}[value]

    @model_validator(mode='before')
    @classmethod
    def nest_aircraft(cls, data: dict) -> dict:
        if not isinstance(data.get('aircraft'), dict):
            data['aircraft'] = dict(aircraft_type_id=data.pop('aircraft_type_id', None),
                                    aircraft_type_name=data.pop('aircraft_type_name', None))
        return data

    @field_validator('takeoff_et', 'at_other_et', mode='before')
    @classmethod
    def remove_invalid_dates(cls, value: str | None) -> datetime | None:
        try:
            datetime.fromisoformat(value)
            return value
        except (ValueError, TypeError):
            return


class Schedule(Base):
    flights: list[Flight] = Field(default_factory=list, alias='items')



import urllib.parse
import pydantic
from typing import Literal, Any
from aiohttp import ClientSession
from datetime import datetime
from logging import getLogger
from functools import cached_property

from . import (
    urls,
    models,
    utils,
)

svo_logger = getLogger('svo-api')


class AsyncSvoApi:
    def __init__(self, session: ClientSession | None = None) -> None:
        self._session = session

    @cached_property
    def session(self):
        if self._session is None:
            self._session = ClientSession()
        return self._session

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.session.close()

    async def _request(self, url: str, params: dict, _session: ClientSession | None = None, **kwargs: Any) -> dict:
        response = await (_session or self.session).get(url, params=params, **kwargs)
        response.raise_for_status()
        return await response.json()

    async def get_schedule(self,
                           direction: Literal['arrival', 'departure'],
                           date_start: datetime,
                           date_end: datetime,
                           per_page: int = 99999,
                           page: int = 0,
                           locale: str = 'ru',
                           **kwargs
                           ) -> models.Schedule:

        response = await self._request(
            url=urls.timetable,
            params=dict(
                direction=direction,
                dateStart=utils.format_date(date_start),
                dateEnd=utils.format_date(date_end),
                perPage=per_page,
                page=page,
                locale=locale,
            ),
            **kwargs,
        )

        schedule = models.Schedule()

        for item in response['items']:
            try:
                flight = models.Flight.model_validate(item)
                schedule.flights.append(flight)
            except pydantic.ValidationError as err:
                svo_logger.warning(f'Skip flight id={item.get("i_id")}\n' + str(err))

        return schedule

    async def get_flight(self,
                         flight_id: int,
                         locale: str = 'ru',
                         **kwargs
                         ) -> models.Flight:

        response = await self._request(
            url=urllib.parse.urljoin(urls.timetable, f'{flight_id}/'),
            params=dict(
                locale=locale,
            ),
            **kwargs
        )
        return models.Flight.model_validate(response)



from unittest import TestCase
from src.aero_svo_api import models
from . import payload


class TestModels(TestCase):
    def test_aircraft_on_valid(self):
        valid_data = payload.AircraftPayload().model_dump()
        aircraft = models.Aircraft.model_validate(valid_data)

    def test_country_on_valid(self):
        valid_data = payload.CountryPayload().model_dump()
        country = models.Country.model_validate(valid_data)

    def test_city_on_valid(self):
        valid_data = payload.CityPayload().model_dump()
        country = models.City.model_validate(valid_data)

    def test_airport_on_valid(self):
        valid_data = payload.AirportPayload().model_dump()
        airport = models.Airport.model_validate(valid_data)

    def test_company_on_valid(self):
        valid_data = payload.CompanyPayload().model_dump()
        company = models.Company.model_validate(valid_data)

    def test_flight_on_valid(self):
        valid_data = payload.FlightPayload().model_dump()
        flight = models.Flight.model_validate(valid_data)
